/*
 * Hytec PCI 5992 Device Driver
 *
 * Copyright (C) 2005 CCLRC
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/proc_fs.h>

#include <linux/pci.h>
#include <linux/fs.h>
#include <linux/mm.h>

#include <asm/uaccess.h>
#include <asm/io.h>

#include "pci5992.h"


MODULE_AUTHOR("CCLRC Daresbury Laboratory");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("PCI 5992 Device Driver");
MODULE_SUPPORTED_DEVICE("PCI-5992");

MODULE_PARM(debug_level, "i");


#ifndef CONFIG_PCI
#error "This driver requires that PCI support be configured in the Kernel"
#endif


static struct file_operations pci5992_fops = {
        owner:          THIS_MODULE,
        ioctl:          pci5992_ioctl,
        open:           pci5992_open,
        release:        pci5992_release,
};

static struct pci_device_id pci5992_idtbl[] __initdata = {
        { PCI_VENDOR_ID_PCI5992, PCI_DEVICE_ID_PCI5992, PCI_ANY_ID, PCI_ANY_ID, PCI_CLASS_OTHERS, 0, 0 },
        { 0, }, /* terminate list */
};
MODULE_DEVICE_TABLE(pci, pci5992_idtbl);

static struct pci_driver pci5992_driver = {
    name: "pci5992", 
    id_table: pci5992_idtbl,
    probe: pci5992_init_board,
    remove: pci5992_remove_board,
};


static const char *driver_name = "pci5992";
static unsigned int pci5992_major = PCI5992_MAJOR;


#ifdef CONFIG_PROC_FS
static struct proc_dir_entry *pci5992_proc_dir_entry = NULL;
static const char* proc_dir_entry = "driver/pci5992";
#endif


/*    Only one PCI5992 board is supported so the device instance block follows inline    */

struct pci_dev *pdev = NULL;

/*    module hardware registers  */

unsigned long badr0_iobase;    /* BADR0 I/O address - unmapped */
size_t badr0_len;              /* BADR0 size in bytes */
void * badr0_mapped_iobase;    /* BADR0 base address - mapped  */

unsigned long badr1_iobase;    /* BADR1 I/O address - unmapped */
size_t badr1_len;              /* BADR1 size in bytes */
void * badr1_mapped_iobase;    /* BADR1 base address - mapped  */


/*    CAMAC data   */

u32 cnaf = 0;
u32 data = 0;
u32 qx = 0;

/*    module STATUS data  */

u32 lamstatus = 0;
u32 serialstatus = 0;
u32 operationstatus = 0;


static void execCAMAC ()
{
    int c, n, a, f;
    int timedOut;

    lamstatus = 0;
    serialstatus = 0;
    operationstatus = 0;

    c = (cnaf >> 16) & 0x3f;
    n = (cnaf >> 9) & 0x1f;
    a = (cnaf >> 5) & 0xf;
    f = cnaf & 0x1f;

    outl(0, badr1_iobase+PCI5992_LAMSTATUS);     //    clear LAM status by clearing bit 12

    outl(cnaf, badr1_iobase+PCI5992_CNAF);       //    write the CNAF code to the module

//    if the CAMAC operation does not require data (read or control operation) the CAMAC cycle starts now

    if (f >= 16 && f <= 23) {
        outl(data, badr1_iobase+PCI5992_DATA);          //    a write  - need data
        DPRINTF("C=%d, N=%d, A=%d, F=%d, data=0x%x\n",c ,n ,a ,f ,data);
    } else {
        data = 0;
        DPRINTF("C=%d, N=%d, A=%d, F=%d\n",c ,n ,a ,f);
    }

//    for CAMAC write operations the CAMAC cycle starts when we have written the data

//    loop waiting for the cycle to complete  -  check SHR bit  (bit 8)  in the LAM STATUS register 

    timedOut = TIMEOUT;
    while ((inl(badr1_iobase+PCI5992_LAMSTATUS) & 0x0100) == 0 && timedOut-- > 0) {continue;}

    if (timedOut == 0) {                      //    waiting loop has timed out 

//    initialise the module

        outl(1, badr1_iobase+PCI5992_STATUS);
        outl(0x1000, badr1_iobase+PCI5992_LAMMASK);
        outl(0, badr1_iobase+PCI5992_LAMSTATUS);

        DPRINTF("CAMAC operation time out\n");
        qx = 4;
    }  else {

//    obtain Q and X

        qx = (inl(badr1_iobase+PCI5992_STATUS) >> 2) & 3;
        DPRINTF("Q=%d, X=%d\n", (qx>>1)&1, qx&1);

//    for CAMAC read operations the return data is now available
    
        if (f <= 7) {
            data = inl(badr1_iobase+PCI5992_DATA);          //    a read - fetch data
            DPRINTF("C=%d, N=%d, A=%d, F=%d, data=0x%x\n",c ,n ,a ,f ,data);
        }
    }

//    record status registers 

    operationstatus = inl(badr1_iobase+PCI5992_STATUS);
    lamstatus = inl(badr1_iobase+PCI5992_LAMSTATUS);
    serialstatus = inl(badr1_iobase+PCI5992_SERIALSTATUS);


}

    
    
/*
   pci5992_init - module initialization
   entered by the kernel when the module is loaded
*/

static int __init  pci5992_init(void) {
    int rc = 0;

    DPRINTF("pci5992_init entered\n");

    rc = register_chrdev(pci5992_major, driver_name, &pci5992_fops);
    if (rc < 0) {
        printk(KERN_WARNING PCI5992
        "In function pci5992_init:\n"
        "register_chrdev() failure - can't get major %d\n", pci5992_major);

        return(rc);
    }
    if (pci5992_major == 0) pci5992_major = rc;

    DPRINTF("register_chrdev returned major = %d\n", pci5992_major);

    rc = pci_module_init(&pci5992_driver);
    if (rc) {
        printk(KERN_WARNING PCI5992
        "In function pci5992_init:\n"
        "pci_module_init() failure - can't find device %d %d\n", PCI_VENDOR_ID_PCI5992, PCI_DEVICE_ID_PCI5992);

        pci_unregister_driver(&pci5992_driver);
        unregister_chrdev(pci5992_major, driver_name);
        return (rc);
    }
 
#ifdef CONFIG_PROC_FS

    pci5992_proc_dir_entry = create_proc_read_entry(proc_dir_entry, 0, NULL, pci5992_read_procmem, NULL);
    if (pci5992_proc_dir_entry == NULL)
        printk(KERN_WARNING PCI5992
            "In function pci5992_init:"
            "create_proc_read_entry() failed"
            " - no /proc file system entry will be available"
            " for the %s driver\n", driver_name );
#endif

    DPRINTF("pci5992_init complete\n");

    return(0);
}


/*
 * pci5992_cleanup - module cleanup
 * entered by the kernel when the module is unloaded
*/

static void __exit pci5992_exit(void)
{

    DPRINTF("pci5992 module is exiting\n");

    pci_unregister_driver(&pci5992_driver);
    unregister_chrdev(pci5992_major, driver_name);

#ifdef CONFIG_PROC_FS
    if (pci5992_proc_dir_entry) remove_proc_entry(proc_dir_entry, NULL);
#endif

}


/*    pci5992_open      -   open method    */

static int pci5992_open(struct inode* inode, struct file* filp)
{

/*    claim all required resources   */

    DPRINTF("pci5992_open:\n");

    MOD_INC_USE_COUNT;     /* increment kernel usage count */

    return(0);
}


/*    pci5992_release      -   release method    */

static int pci5992_release(struct inode* inode, struct file* filp)
{

/*    free all resources   */

    DPRINTF("pci5992_release:\n");

    MOD_DEC_USE_COUNT;     /* decrement kernel usage count */

    return(0);
}


/*   pci5992_ioctl - ioctl method     */

static int pci5992_ioctl(struct inode *inode, struct file *filp, unsigned int cmd,
         unsigned long arg)
{

    DPRINTF("pci5992_ioctl: called with command %d\n", cmd);

    switch (cmd) {
        case DEFINE_CNAF:
            if (copy_from_user(&cnaf, (void *) arg, sizeof (cnaf))) {
                return -EFAULT;
            }
            break;
        case DEFINE_DATA:
            if (copy_from_user(&data, (void *) arg, sizeof (data))) {
                return -EFAULT;
            }
            break;
        case ACQUIRE_DATA:
            if (copy_to_user((void *) arg, &data, sizeof (data))) {
                return -EFAULT;
            }
            break;
        case ACQUIRE_QX:
            if (copy_to_user((void *) arg, &qx, sizeof (qx))) {
                return -EFAULT;
            }
            break;
        case ACQUIRE_LAMSTATUS:
            if (copy_to_user((void *) arg, &lamstatus, sizeof (lamstatus))) {
                return -EFAULT;
            }
            break;
        case ACQUIRE_SERIALSTATUS:
            if (copy_to_user((void *) arg, &serialstatus, sizeof (serialstatus))) {
                return -EFAULT;
            }
            break;
        case ACQUIRE_OPERATIONSTATUS:
            if (copy_to_user((void *) arg, &operationstatus, sizeof (operationstatus))) {
                return -EFAULT;
            }
            break;
        case EXECUTE:
            execCAMAC ();
            break;
        default:
            return -ENOTTY;
    }

    return(0);

}


/*   pci5992_init_board  - init each board as we find them      */

static int __init pci5992_init_board(struct pci_dev *pdev, const struct pci_device_id *ent)
{
    int rc;
    u16 word;
    u32 data;

    DPRINTF("pci5992_init_board:\n");
   
    rc = pci_enable_device(pdev);

    pci_set_master(pdev);      /* set bus master enable */
   
   
    badr0_iobase = pci_resource_start(pdev, 0);
    if (!badr0_iobase) {
        printk(KERN_ERR PCI5992
        "In function pci5992_init_board:\n"
        "pci_resource_start() failure - PCI Region 0 unavailable\n");
//        return(-ENODEV);
    }

    badr0_len = pci_resource_len(pdev, 0);
    if (!badr0_len) {
        printk(KERN_ERR PCI5992
        "In function pci5992_init_board:\n"
        "pci_resource_len() reports PCI Region 0 of size 0 bytes\n");
//        return(-ENODEV);
    }
    
    if (!request_region (badr0_iobase, badr0_len, driver_name)) {
        printk(KERN_ERR PCI5992
        "In function pci5992_init_board:\n"
        "request_region() reports PCI Region 0 unavailable\n");
//                return(-ENODEV);
    }

    DPRINTF("pci address space 0 at: 0x%lx length %d\n", badr0_iobase, badr0_len);
    
    badr1_iobase = pci_resource_start(pdev, 1);
    if (!badr1_iobase) {
        printk(KERN_ERR PCI5992
        "In function pci5992_init_board:\n"
        "pci_resource_start() failure - PCI Region 1 unavailable\n");
            release_region(badr0_iobase, badr0_len);
//        return(-ENODEV);
    }

    badr1_len = pci_resource_len(pdev, 1);
    if (!badr1_len) {
        printk(KERN_ERR PCI5992
        "In function pci5992_init_board:\n"
        "pci_resource_len() reports PCI Region 1 of size 0 bytes\n");
            release_region(badr0_iobase, badr0_len);
//        return(-ENODEV);
    }
    
    if (!request_region (badr1_iobase, badr1_len, driver_name)) {
        printk(KERN_ERR PCI5992
        "In function pci5992_init_board:\n"
        "request_region() reports PCI Region 1 unavailable\n");
            release_region(badr0_iobase, badr0_len);
//                return(-ENODEV);
    }
    
    DPRINTF("pci address space 1 at: 0x%lx length %d\n", badr1_iobase, badr1_len);
        
    rc = pci_read_config_word(pdev, 0, &word);
    DPRINTF("pci_read_config_word 0 returned %d 0x%x\n", rc, word);
    rc = pci_read_config_word(pdev, 2, &word);
    DPRINTF("pci_read_config_word 2 returned %d 0x%x\n", rc, word);

//    initialise the module

    outl(1, badr1_iobase+PCI5992_STATUS);
    outl(0x1000, badr1_iobase+PCI5992_LAMMASK);
    outl(0, badr1_iobase+PCI5992_LAMSTATUS);

//    and save the status 

    data = inw(badr1_iobase+PCI5992_STATUS);
    DPRINTF("status= 0x%x\n", data);
    operationstatus = data;

    data = inw(badr1_iobase+PCI5992_LAMSTATUS);
    DPRINTF("lam status= 0x%x\n", data);
    lamstatus = data;

    data = inw(badr1_iobase+PCI5992_SERIALSTATUS);
    DPRINTF("serial status= 0x%x\n", data);
    serialstatus = data;

    return(0);
}


/*   pci5992_remove_board  -  remove allocated resources  */

static void __exit pci5992_remove_board(struct pci_dev *pdev) {

    DPRINTF("pci_remove_board:\n");
   
    release_region(badr0_iobase, badr0_len);
    release_region(badr1_iobase, badr1_len);
       
}


//    define module initialisation and termination procedures
module_init(pci5992_init);
module_exit(pci5992_exit);

#ifdef CONFIG_PROC_FS
static int pci5992_read_procmem(char *buf, char **start, off_t offset, int count,
            int *eof, void *data)
{

        return(0);
}
#endif
